# from tornado.web import RequestHandler
#
#
# class BaseClient(RequestHandler):
#
#     def get(self, *args, **kwargs):
#         self.write("say something")
#
#     def set_default_headers(self, *args, **kwargs):
#         self.set_header("Access-Control-Allow-Origin", "*")
#         self.set_header("Access-Control-Allow-Headers", "x-requested-with")
#         self.set_header("Access-Control-Allow-Methods", "POST, GET, OPTIONS")
